#include <stdio.h>


#define SIGN_POS 0
#define SIGN_NEG 0x80
#define RC_RND 0
#define RC_DOWN 0x400
#define RC_UP 0x800
#define RC_CHOP 0xc00

typedef unsigned short u16;

typedef struct fpu__reg {
  unsigned sigl;
  unsigned sigh;
  short exp;
} FPU_REG;


int FPU_round(FPU_REG *x, unsigned long extent, int dummy,
	      unsigned int control_w, unsigned char sign);

#define PR_64_BITS 0x300
#define PR_53_BITS 0x200
#define PR_24_BITS 0x000


#define FULL_PRECISION (PR_64_BITS | RC_RND | 0x3f)


void test_64_all_rnd(FPU_REG *x)
{
  FPU_REG y;
  int tag, i, j, k, m;
  int round[4] = { RC_RND, RC_CHOP, RC_DOWN, RC_UP };
  int sign[2] = { SIGN_POS, SIGN_NEG };
  int mask[2] = { 0, 0x3f };
  unsigned long extent[4] = { 0x80000000, 1, 0xffffffff, 0 };
  long double *d = (long double *)&y;

  for ( m = 0; m < 2; m++ )
    {
      for ( k = 0; k < 4; k++ )
	{
	  for ( j = 0; j < 2; j++ )
	    {
	      for ( i = 0; i < 4; i++ )
		{
		  y = *x;
		  tag = FPU_round(&y, extent[k], 0,
				  (PR_64_BITS | round[i] | mask[m]),
				  sign[j]);
		  printf("%d %d:%d:%d %04x %04x %08lx %08lx %d %Lg\n",
			 64, i, j, k, PR_64_BITS | round[i] | mask[m],
			 (u16)y.exp, y.sigh, y.sigl, tag, *d);
		}
	    }
	}
    }

}


void test_53_all_rnd(FPU_REG *x)
{
  FPU_REG y;
  int tag, i, j, k;
  int round[4] = { RC_RND, RC_CHOP, RC_DOWN, RC_UP };
  int sign[2] = { SIGN_POS, SIGN_NEG };
  unsigned long extent[4] = { 0, 1, 0xffffffff, 0 };
  unsigned long sigl[4] = { 0xc00, 0x800, 0xfff, 0x800 };

  for ( k = 0; k < 4; k++ )
    {
      for ( j = 0; j < 2; j++ )
	{
	  for ( i = 0; i < 4; i++ )
	    {
	      y = *x;
	      y.sigl &= 0xfffff800;
	      y.sigl |= sigl[k];
	      tag = FPU_round(&y, extent[k], 0,
			      (PR_53_BITS | round[i] | 0x3f),
			      sign[j]);
	      printf("%d %d:%d:%d %04x %08lx %08lx %d\n",
		     53, i, j, k,
		     (u16)y.exp, y.sigh, y.sigl, tag);
	    }
	}
    }

}


void test_24_all_rnd(FPU_REG *x)
{
  FPU_REG y;
  int tag, i, j, k;
  int round[4] = { RC_RND, RC_CHOP, RC_DOWN, RC_UP };
  int sign[2] = { SIGN_POS, SIGN_NEG };
  unsigned long extent[4] = { 0, 1, 0xffffffff, 0 };
  unsigned long sigh[4] = { 0x180, 0x100, 0x1ff, 0x100 };

  for ( k = 0; k < 4; k++ )
    {
      for ( j = 0; j < 2; j++ )
	{
	  for ( i = 0; i < 4; i++ )
	    {
	      y = *x;
	      y.sigl = 0;
	      y.sigh &= 0xffffff00;
	      y.sigh |= sigh[k];
	      tag = FPU_round(&y, extent[k], 0,
			      (PR_24_BITS | round[i] | 0x3f),
			      sign[j]);
	      printf("%d %d:%d:%d %04x %08lx %08lx %d\n",
		     24, i, j, k,
		     (u16)y.exp, y.sigh, y.sigl, tag);
	    }
	}
    }

}


int main()
{
  int tag;
  FPU_REG a = { 0x1b1b1b1b, 0xa3a3a3a3, 0 };
  FPU_REG b = { 0xffffffff, 0xffffffff, 0x3fff };
  FPU_REG c = { 0xffffffff, 0xffffffff, -0x3fff };
  FPU_REG d = { 0xffffffff, 0xffffffff, 0x3fff };
  FPU_REG e = { 0xffffffff, 0xffffffff, 0x4100 };
  FPU_REG x;

  test_64_all_rnd(&a);
  test_53_all_rnd(&a);
  test_24_all_rnd(&a);

  test_64_all_rnd(&b);
  test_53_all_rnd(&b);
  test_24_all_rnd(&b);

  test_64_all_rnd(&c);
  test_64_all_rnd(&d);
  test_64_all_rnd(&e);

}


void FPU_exception(int n)
{

  printf("FPU_exception(%04x)\n", n);

}

void set_precision_flag_down() { printf("precision flag down\n"); }
void set_precision_flag_up() { printf("precision flag up\n"); }
int arith_overflow() { printf("arith_overflow()\n"); return 1; }

